@@ -20,11 +20,10 @@ def validate_properties(data: Dict, context: str):
 def _get_keys(data: Union[Dict, List]) -> List[str]:
     """
     Recursively walk a dictionary or list collecting keys.
-    Ignore dictionaries mapped to the 'environment' and 'dockerEnvironment'
-    keys as these will contain user-specified keys.
+    Exclude dictionaries with user-specified keys.
     """
     keys: List[str] = []
-    excluded_keys = [ENV, DOCKER_ENV]
+    excluded_keys = [ENV, DOCKER_ENV, MVARS]
     if isinstance(data, dict):
         for key, value in data.items():
             keys.append(key)
